// SPDX-License-Identifier: BSD-3-Clause

#include "osmem.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <unistd.h>
#include <sys/mman.h>
#include "block_meta.h"

#define PAGE_SIZE 8
#define ALIGN_TO_PAGE(size) (((size) + (PAGE_SIZE - 1)) & ~(PAGE_SIZE - 1))
#define BLOCK_META_SIZE_PADDED (ALIGN_TO_PAGE(sizeof(struct block_meta)))
#define MMAP_THRESHOLD (1 << 17)

#define MAP_ANONYMOUS 0x20
#define MEM_PAGE_SIZE (1<<12)

void *memory_list_head;
unsigned long limit;

void *ptr_from(struct block_meta *block)
{
	return (void *)((char *)block + BLOCK_META_SIZE_PADDED);
}

struct block_meta *block_from(void *ptr)
{
	return (struct block_meta *)((char *)ptr - BLOCK_META_SIZE_PADDED);
}

/*
 * @note block->size = actuall size plus padd
 *
 */
void init_memory_list(void)
{
	struct block_meta *list_head = (struct block_meta *) memory_list_head;

	list_head->next = NULL;
	list_head->prev = NULL;
	list_head->size = MMAP_THRESHOLD - BLOCK_META_SIZE_PADDED;
	list_head->status = STATUS_FREE;
}

/*
 * @brief prealocates MMAP_THRESHOLD bytes
 * @pre malloc is called first time ever for size <= MMAP_THRESHOLD
 * @note prealloc will always prealloc MMAP_THRESHOLD bytes
 */
void prealloc(void)
{
	memory_list_head = sbrk(MMAP_THRESHOLD);
	init_memory_list();
}

/*
 * @brief Find best FREE block on the heap that has size >= than my requested size
 *
 * @return NULL if no free block was found
 *
 * struct block_meta *block that meets the requirement
 */
struct block_meta *find_best_free_block(size_t size)
{
	struct block_meta *list_head = memory_list_head;
	struct block_meta *best_block = NULL;

	while (list_head) {
		if (list_head->size >= ALIGN_TO_PAGE(size) && list_head->status == STATUS_FREE) {
			if (!best_block)
				best_block = list_head;

			if (best_block->size - ALIGN_TO_PAGE(size) >
				list_head->size - ALIGN_TO_PAGE(size)) {
				best_block = list_head;
			}
		}
		list_head = list_head->next;
	}

	return best_block;
}

/*
 * @brief Truncate a large heap block to new_size, remaining block is marked FREE
 *
 * If there is at least one byte to be used later the function truncates the block,
 * otherwise the whole block is returned, because it couldn't be trunctated.
 *
 * @pre Function must be called for STATUS_FREE or STATUS_ALLOC only
 *
 * @note This function never fails if called.
 * @note Address to block does not change.
 *
 * @return struct block_meta *block (the one given to the function)
 */
struct block_meta *truncate_block(struct block_meta *block, size_t new_size)
{
	size_t old_size = block->size;
	struct block_meta *last = block->next;

	if (old_size -  ALIGN_TO_PAGE(new_size) >= ALIGN_TO_PAGE(BLOCK_META_SIZE_PADDED + 1)) {
		block->size = ALIGN_TO_PAGE(new_size);

		block->next = (struct block_meta *)((char *)block + block->size + BLOCK_META_SIZE_PADDED);

		block->next->next = last;
		block->next->prev = block;
		block->next->status = STATUS_FREE;
		block->next->size = old_size - block->size - BLOCK_META_SIZE_PADDED;

		if (last)
			last->prev = block->next;
	}

	return block;
}

struct block_meta *get_list_tail(void)
{
	struct block_meta *list_head = (struct block_meta *) memory_list_head;

	if (!list_head)
		return NULL;

	while (list_head->next)
		list_head = list_head->next;

	return list_head;
}

/*
 * @brief Call mmap or sbrk
 *
 * If size is smaller than MMAP_THRESHOLD, sbrk is called,
 * adding the new block to the linked list.
 *
 * If size is bigger than MMAP_THRESHOLD,
 * mmap is called and the node is not added to the list
 *
 * @return pointer to the usable memory
 */
void *request_memory(size_t size)
{
	// try mmap
	if (ALIGN_TO_PAGE(size + BLOCK_META_SIZE_PADDED) > limit) {
		struct block_meta *os_block = (struct block_meta *) mmap(NULL, ALIGN_TO_PAGE(size + BLOCK_META_SIZE_PADDED),
																PROT_READ | PROT_WRITE, MAP_PRIVATE | MAP_ANONYMOUS, -1, 0);

		if ((void *)os_block == (void *) -1) {
			errno = ENOMEM;
			return NULL;
		}

		os_block->next = NULL;
		os_block->prev = NULL;
		os_block->size = ALIGN_TO_PAGE(size);
		os_block->status = STATUS_MAPPED;

		return ptr_from(os_block);
	}

	// mmap not applicable so i try sbrk
	struct block_meta *memory_list_tail = get_list_tail();
	struct block_meta *memory_list_new_tail = (struct block_meta *)sbrk(ALIGN_TO_PAGE(size + BLOCK_META_SIZE_PADDED));

	if ((void *)memory_list_new_tail == (void *) -1) {
		errno = ENOMEM;
		return NULL;
	}

	memory_list_new_tail->next = NULL;
	memory_list_new_tail->prev = memory_list_tail;
	memory_list_new_tail->size = ALIGN_TO_PAGE(size);
	memory_list_new_tail->status = STATUS_ALLOC;

	memory_list_tail->next = memory_list_new_tail;

	return ptr_from(memory_list_new_tail);
}

/*
 * @brief Allocate memory on heap or somewhere else with mmap
 *
 * If size <= MMAP_THRESHOLD we do:
 *	1. find first free block;
 *	2. try truncate it;
 *	3. try expand last block if no free block was found;
 *	4. everything failed, request memory
 *
 * If size > MMAP_THRESHOLD we do:
 *	1. request size
 *
 * @return void *usable_memory
 */
void *os_malloc(size_t size)
{
	/* TODO: Implement os_malloc */
	limit = MMAP_THRESHOLD;

	if (size == 0)
		return NULL;

	if (memory_list_head == NULL && ALIGN_TO_PAGE(size + BLOCK_META_SIZE_PADDED) <= limit)
		prealloc();

	// preallocation was made beforehand
	if (ALIGN_TO_PAGE(size + BLOCK_META_SIZE_PADDED) <= limit) {
		struct block_meta *block = find_best_free_block(size);

		// if such block exists, i truncate it
		if (block) {
			block->status = STATUS_ALLOC;
			truncate_block(block, size);
			return ptr_from(block);
		}
			// try expand the last heap block
			struct block_meta *list_tail = get_list_tail();

			if (list_tail && list_tail->status == STATUS_FREE && ALIGN_TO_PAGE(size) - list_tail->size <= limit) {
				void *check = sbrk(ALIGN_TO_PAGE(size) - list_tail->size);

				if (check == (void *)-1) {
					errno = ENOMEM;
					return NULL;
				}

				list_tail->size = ALIGN_TO_PAGE(size);
				list_tail->status = STATUS_ALLOC;

				return ptr_from(list_tail);
			}
	}

	return request_memory(size);
}

/*
 * @brief Joins two adjaicent blocks
 * @note Function always merges right block to the left one,
 * regardless of status
 * @return void
 */
void join_blocks(struct block_meta *left)
{
	struct block_meta *right = left->next->next;
	struct block_meta *middle = left->next;

	if (right)
		right->prev = left;

	left->next = right;
	left->size += middle->size + BLOCK_META_SIZE_PADDED;
}

/*
 * @brief Joins neighbouring free blocks
 */
void coalesce(void)
{
	struct block_meta *head = memory_list_head;

	// no coalesce for a single block or NULL
	if (!head || !head->next)
		return;

	head = head->next;

	while (head) {
		if (head->status == STATUS_FREE && head->prev->status == STATUS_FREE) {
			struct block_meta *first = head->prev;

			join_blocks(first);
			head = first;
		}

		head = head->next;
	}
}

void os_free(void *ptr)
{
	/* TODO: Implement os_free */
	if (ptr == NULL)
		return;

	struct block_meta *block = block_from(ptr);

	if (block->status == STATUS_ALLOC) {
		block->status = STATUS_FREE;
		coalesce();
	}

	// am bloc mmap-uit
	if (block->status == STATUS_MAPPED)
		munmap((void *)block, block->size + BLOCK_META_SIZE_PADDED);
}

/*
 * @brief malloc version called by calloc only
 * @note All malloc logic applyes here except
 * some key parts where MMAP_THRESHOLD is always used
 */
void *os_malloc_calloc(size_t size)
{
	/* TODO: Implement os_malloc */
	limit = MEM_PAGE_SIZE;

	if (size == 0)
		return NULL;

	if (memory_list_head == NULL && ALIGN_TO_PAGE(size + BLOCK_META_SIZE_PADDED) <= limit)
		prealloc();

	// preallocation was made beforehand
	if (ALIGN_TO_PAGE(size + BLOCK_META_SIZE_PADDED) <= limit) {
		struct block_meta *block = find_best_free_block(size);

		// if such block exists, i truncate it
		if (block) {
			block->status = STATUS_ALLOC;
			truncate_block(block, size);
			return ptr_from(block);
		}
			// try expand the last heap block
			struct block_meta *list_tail = get_list_tail();

			if (list_tail && list_tail->status == STATUS_FREE && ALIGN_TO_PAGE(size) - list_tail->size <= limit) {
				void *check = sbrk(ALIGN_TO_PAGE(size) - list_tail->size);

				if (check == (void *)-1) {
					errno = ENOMEM;
					return NULL;
				}

				list_tail->size = ALIGN_TO_PAGE(size);
				list_tail->status = STATUS_ALLOC;

				return ptr_from(list_tail);
			}
	}

	return request_memory(size);
}

void *os_calloc(size_t nmemb, size_t size)
{
	/* TODO: Implement os_calloc */
	limit = MEM_PAGE_SIZE;
	void *rez = os_malloc_calloc(nmemb * size);

	if (rez == NULL)
		return NULL;

	memset(rez, 0, nmemb * size);
	return rez;
}

/*
 * @brief Expands a block of memory on the heap
 *
 * If the block is inside the heap, an expand is tried.
 * If the block is at the end, an extend is tried.
 *
 * @return NULL on allocation fail or expansion fail,
 * otherwise valid usable pointer
 */
void *expand_block_realloc(struct block_meta *block, size_t size)
{
	if (block->next) {
		// block is most definetly inside the heap, if next block exists
		coalesce();
		if (block->next->status == STATUS_FREE &&
			block->size + block->next->size + BLOCK_META_SIZE_PADDED >= ALIGN_TO_PAGE(size)) {
			join_blocks(block);
			truncate_block(block, size);
			return ptr_from(block);
		}
	}

	if (!block->next) {
		// block is definetly at the end; followed by nothing
		// i think i must check the adding up of the sizes to be less than MMAP_threshold
		void *check = sbrk(ALIGN_TO_PAGE(size) - block->size);

		if (check == (void *)-1) {
			errno = ENOMEM;
			return NULL;
		}

		block->size = ALIGN_TO_PAGE(size);
		return ptr_from(block);
	}

	// operation of expanding has failed
	return NULL;
}

/*
 * @brief realloc a pointer
 *
 * If ptr = null => malloc (size)
 * If size = 0 => free(ptr)
 * If status = FREE => return NULL
 *
 * If block is on heap:
 *	1. new_size = old_size
 *	2. new_size < old_size
 *	3. new_size > old_size
 *
 * Block not on heap or everything failed
 */
void *os_realloc(void *ptr, size_t size)
{
	/* TODO: Implement os_realloc */
	limit = MMAP_THRESHOLD;

	if (ptr == NULL)
		return os_malloc(size);

	if (size == 0) {
		os_free(ptr);
		return NULL;
	}

	struct block_meta *block = block_from(ptr);

	if (block->status == STATUS_FREE)
		return NULL;

	if (block->status == STATUS_ALLOC) {
		if (ALIGN_TO_PAGE(size) == block->size)
			return ptr;

		if (ALIGN_TO_PAGE(size) < block->size) {
			truncate_block(block, size);
			return ptr_from(block);
		}

		if (ALIGN_TO_PAGE(size) > block->size) {
			void *check = expand_block_realloc(block, size);

			if (check)
				return check;
		}
	}

	// realloc part
	void *new_ptr = os_malloc(size);

	if (new_ptr) {
		memcpy(new_ptr, ptr, block->size < ALIGN_TO_PAGE(size) ? block->size : ALIGN_TO_PAGE(size));
		os_free(ptr);
		return new_ptr;
	}

	return NULL;
}
